@@ -3,30 +3,21 @@ import re
 import sys
 import textwrap
 import typing
-from collections import defaultdict
 from typing import (
-    Any, AnyStr, Callable, Dict, Generic, Mapping, NewType, Optional, Pattern,
-    Tuple, TypeVar, Union, Type)
+    Any, AnyStr, Callable, Dict, Generic, Mapping, NewType, Optional, Pattern, Match, Tuple,
+    TypeVar, Union, Type)
 
 import pytest
 import typing_extensions
 
-from sphinx_autodoc_typehints import format_annotation, process_docstring
+from sphinx_autodoc_typehints import (
+    format_annotation, process_docstring, get_annotation_module, get_annotation_class_name,
+    get_annotation_args)
 
 try:
-    from typing import ClassVar  # not available prior to Python 3.5.3
+    from typing import IO
 except ImportError:
-    ClassVar = None
-
-try:
-    from typing import NoReturn  # not available prior to Python 3.6.5
-except ImportError:
-    NoReturn = None
-
-try:
-    from typing import Literal
-except ImportError:
-    Literal = defaultdict(lambda: None)
+    from typing.io import IO
 
 T = TypeVar('T')
 U = TypeVar('U', covariant=True)
@@ -38,9 +29,13 @@ class A:
     def get_type(self):
         return type(self)
 
+    class Inner:
+        pass
+
 
 class B(Generic[T]):
-    pass
+    # This is set to make sure the correct class name ("B") is picked up
+    name = 'Foo'
 
 
 class C(B[str]):
@@ -63,6 +58,43 @@ class Metaclass(type):
     pass
 
 
+@pytest.mark.parametrize('annotation, module, class_name, args', [
+    pytest.param(str, 'builtins', 'str', (), id='str'),
+    pytest.param(None, 'builtins', 'None', (), id='None'),
+    pytest.param(Any, 'typing', 'Any', (), id='Any'),
+    pytest.param(AnyStr, 'typing', 'AnyStr', (), id='AnyStr'),
+    pytest.param(Dict, 'typing', 'Dict', (), id='Dict'),
+    pytest.param(Dict[str, int], 'typing', 'Dict', (str, int), id='Dict_parametrized'),
+    pytest.param(Dict[T, int], 'typing', 'Dict', (T, int), id='Dict_typevar'),
+    pytest.param(Tuple, 'typing', 'Tuple', (), id='Tuple'),
+    pytest.param(Tuple[str, int], 'typing', 'Tuple', (str, int), id='Tuple_parametrized'),
+    pytest.param(Union[str, int], 'typing', 'Union', (str, int), id='Union'),
+    pytest.param(Callable, 'typing', 'Callable', (), id='Callable'),
+    pytest.param(Callable[..., str], 'typing', 'Callable', (..., str), id='Callable_returntype'),
+    pytest.param(Callable[[int, str], str], 'typing', 'Callable', (int, str, str),
+                 id='Callable_all_types'),
+    pytest.param(Pattern, 'typing', 'Pattern', (), id='Pattern'),
+    pytest.param(Pattern[str], 'typing', 'Pattern', (str,), id='Pattern_parametrized'),
+    pytest.param(Match, 'typing', 'Match', (), id='Match'),
+    pytest.param(Match[str], 'typing', 'Match', (str,), id='Match_parametrized'),
+    pytest.param(IO, 'typing', 'IO', (), id='IO'),
+    pytest.param(W, 'typing', 'NewType', (str,), id='W'),
+    pytest.param(Metaclass, __name__, 'Metaclass', (), id='Metaclass'),
+    pytest.param(Slotted, __name__, 'Slotted', (), id='Slotted'),
+    pytest.param(A, __name__, 'A', (), id='A'),
+    pytest.param(B, __name__, 'B', (), id='B'),
+    pytest.param(C, __name__, 'C', (), id='C'),
+    pytest.param(D, __name__, 'D', (), id='D'),
+    pytest.param(E, __name__, 'E', (), id='E'),
+    pytest.param(E[int], __name__, 'E', (int,), id='E_parametrized'),
+    pytest.param(A.Inner, __name__, 'A.Inner', (), id='Inner')
+])
+def test_parse_annotation(annotation, module, class_name, args):
+    assert get_annotation_module(annotation) == module
+    assert get_annotation_class_name(annotation, module) == class_name
+    assert get_annotation_args(annotation, module, class_name) == args
+
+
 @pytest.mark.parametrize('annotation, expected_result', [
     (str,                           ':py:class:`str`'),
     (int,                           ':py:class:`int`'),
@@ -70,12 +102,6 @@ class Metaclass(type):
     (type,                          ':py:class:`type`'),
     (Type,                          ':py:class:`~typing.Type`'),
     (Type[A],                       ':py:class:`~typing.Type`\\[:py:class:`~%s.A`]' % __name__),
-    pytest.param(NoReturn,          ':py:data:`~typing.NoReturn`',
-                 marks=[pytest.mark.skipif(NoReturn is None,
-                                           reason='typing.NoReturn is not available')]),
-    pytest.param(ClassVar[str],      ':py:data:`~typing.ClassVar`\\[:py:class:`str`]',
-                 marks=[pytest.mark.skipif(ClassVar is None,
-                                           reason='typing.ClassVar is not available')]),
     (Any,                           ':py:data:`~typing.Any`'),
     (AnyStr,                        ':py:data:`~typing.AnyStr`'),
     (Generic[T],                    ':py:class:`~typing.Generic`\\[\\~T]'),
@@ -116,16 +142,15 @@ class Metaclass(type):
     (Callable[[T], T],              ':py:data:`~typing.Callable`\\[\\[\\~T], \\~T]'),
     (Pattern,                       ':py:class:`~typing.Pattern`'),
     (Pattern[str],                  ':py:class:`~typing.Pattern`\\[:py:class:`str`]'),
-    pytest.param(Literal['a', 1],   ":py:data:`~typing.Literal`\\['a', 1]",
-                 marks=[pytest.mark.skipif(isinstance(Literal, defaultdict),
-                                           reason='Requires Python 3.8+')]),
+    (IO,                            ':py:class:`~typing.IO`'),
+    (IO[str],                       ':py:class:`~typing.IO`\\[:py:class:`str`]'),
     (Metaclass,                     ':py:class:`~%s.Metaclass`' % __name__),
     (A,                             ':py:class:`~%s.A`' % __name__),
-    (B,                             ':py:class:`~%s.B`\\[\\~T]' % __name__),
+    (B,                             ':py:class:`~%s.B`' % __name__),
     (B[int],                        ':py:class:`~%s.B`\\[:py:class:`int`]' % __name__),
     (C,                             ':py:class:`~%s.C`' % __name__),
     (D,                             ':py:class:`~%s.D`' % __name__),
-    (E,                             ':py:class:`~%s.E`\\[\\~T]' % __name__),
+    (E,                             ':py:class:`~%s.E`' % __name__),
     (E[int],                        ':py:class:`~%s.E`\\[:py:class:`int`]' % __name__),
     (W,                             ':py:func:`~typing.NewType`\\(:py:data:`~W`, :py:class:`str`)')
 ])
@@ -144,19 +169,19 @@ def test_format_annotation(inv, annotation, expected_result):
         m = re.match('^:py:(?P<role>class|data|func):`~(?P<name>[^`]+)`', result)
         assert m, 'No match'
         name = m.group('name')
-        role = next((o.role for o in inv.objects if o.name == name), None)
-        if name in {'typing.Pattern', 'typing.Match', 'typing.NoReturn'}:
-            if sys.version_info < (3, 6):
-                assert role is None, 'No entry in Python 3.5’s objects.inv'
-                return
+        expected_role = next((o.role for o in inv.objects if o.name == name), None)
+        if expected_role:
+            if expected_role == 'function':
+                expected_role = 'func'
 
-        assert role is not None, 'Name {} not found'.format(name)
-        assert m.group('role') == ('func' if role == 'function' else role)
+            assert m.group('role') == expected_role
 
 
 @pytest.mark.parametrize('library', [typing, typing_extensions],
                          ids=['typing', 'typing_extensions'])
 @pytest.mark.parametrize('annotation, params, expected_result', [
+    ('ClassVar', int, ":py:data:`~typing.ClassVar`\\[:py:class:`int`]"),
+    ('NoReturn', None, ":py:data:`~typing.NoReturn`"),
     ('Literal', ('a', 1), ":py:data:`~typing.Literal`\\['a', 1]"),
     ('Type', None, ':py:class:`~typing.Type`'),
     ('Type', (A,), ':py:class:`~typing.Type`\\[:py:class:`~%s.A`]' % __name__)
@@ -188,6 +213,7 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
         sys.path.insert(0, str(test_path))
 
     app.config.always_document_param_types = always_document_param_types
+    app.config.autodoc_mock_imports = ['mailbox']
     app.build()
 
     assert 'build succeeded' in status.getvalue()  # Build succeeded
@@ -196,14 +222,11 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
     warnings = warning.getvalue().strip()
     assert 'Cannot resolve forward reference in type annotations of ' in warnings, warnings
 
+    format_args = {}
     if always_document_param_types:
-        undoc_params = '''
-
-           Parameters:
-              **x** ("int") --'''
-
+        format_args['undoc_params'] = '\n\n   Parameters:\n      **x** ("int") --'
     else:
-        undoc_params = ""
+        format_args['undoc_params'] = ""
 
     text_path = pathlib.Path(app.srcdir) / '_build' / 'text' / 'index.txt'
     with text_path.open('r') as f:
@@ -357,7 +380,7 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
            Return type:
               bytes
 
-        dummy_module.function_with_escaped_default(x='\\x08')
+        dummy_module.function_with_escaped_default(x='\\\\x08')
 
            Function docstring.
 
@@ -438,11 +461,11 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
               Method docstring.
 
               Parameters:
-                 **x** (*Callable**[**[**int**, **bytes**]**, **int**]*) --
+                 **x** ("Optional"["Callable"[["int", "bytes"], "int"]]) --
                  foo
 
               Return type:
-                 ClassWithTypehintsNotInline
+                 "ClassWithTypehintsNotInline"
 
         dummy_module.undocumented_function(x)
 
@@ -456,16 +479,22 @@ def test_sphinx_output(app, status, warning, always_document_param_types):
            Class docstring.
 
            __init__()
-        '''.format(undoc_params=undoc_params)).replace('–', '--')
 
-        if sys.version_info < (3, 6):
-            expected_contents += '''
-      Initialize self.  See help(type(self)) for accurate signature.
-'''
-        else:
-            expected_contents += '''
-      Return type:
-         "None"
-'''
+              Initialize self.  See help(type(self)) for accurate signature.
 
+        @dummy_module.Decorator(func)
+
+           Initializer docstring.
+
+           Parameters:
+              **func** ("Callable"[["int", "str"], "str"]) -- function
+
+        dummy_module.mocked_import(x)
+
+           A docstring.
+
+           Parameters:
+              **x** ("Mailbox") -- function
+        ''')
+        expected_contents = expected_contents.format(**format_args).replace('–', '--')
         assert text_contents == expected_contents
